@@ -164,9 +164,9 @@ def column_to_histo( dfcol , binmax = 32*1024, himax=32*1024, savename = "", hna
     - saves txt
     - updates allhist.root too
     """
-    print("D... ========= creating histo ======= column to_numpy; bins=", binmax)
     narr = dfcol.to_numpy()
-    print("D... len=", len(narr), narr.dtype, "...  ndarray to np histo")
+    print(f"D... ========= creating histo {hname} ======= len={len(narr)}; bins=", binmax)
+    #print("D... len=", len(narr), narr.dtype, "...  ndarray to np histo")
     #print(dfcol)
     #print(narr)
     his,bins = np.histogram(narr,bins=binmax,range=(0.0,himax) )
@@ -244,13 +244,18 @@ def endblock(nomem=False):
         print( freemem() )
     print("D... ","_"*50)
 
+
+
+
+
+
 #------------------------------------------------------------------------
 def select_channels(df, channels, delete_original = False):
     """
     select channels bu not as view! COPY and DELETE
     """
     print("D... selecting channels:", channels)
-    print( freemem() )
+    #print( freemem() )
     # CHANNEL (i hope it is sorted)
     if delete_original:
         df1 = df.loc[ df["ch"].isin( channels ) ].copy()
@@ -266,28 +271,88 @@ def select_channels(df, channels, delete_original = False):
     return df1
 
 
+
+
+
 def enhance_by_dtus_and_next_E(df):
     print("D... adding dtus and next_E columns")
     print( freemem() )
 
-    print(f"D... broadening table - dt")
+
+    print(f"D... broadening table with dt :next event  in t+dtus")
     df['dtus'] = (df.time.shift(-1)-df.time)*1000*1000
-    df.fillna(0)
     df['dtus'] = df['dtus'].astype('float32') #
-    print("D... range of erlang values : ",df['dtus'].min(), df['dtus'].max() )
+    dtusmin = df['dtus'].min()
+    dtusmax = df['dtus'].max()
+    df.fillna(99999, inplace =True)
+    print(f"D... range of time differences (erlang) values : {dtusmin:.3f} us ...  {dtusmax:.3f}  usec." )
+
+
+    print(f"D... broadening table with dt :prev event  = t - dtus (to enable search for standalones)")
+    df['dtuspr'] = (df.time - df.time.shift(+1))*1000*1000
+    df['dtuspr'] = df['dtuspr'].astype('float32') #
+    dtusminpr = df['dtuspr'].min()
+    dtusmaxpr = df['dtuspr'].max()
+    df.fillna(99999, inplace =True)
+    print(f"D... range of time differences (erlang) values : {dtusminpr:.3f} us ...  {dtusmaxpr:.3f}  usec. PREVIOUS" )
+
+    df['dtus3'] = (df.time.shift(-2)-df.time)*1000*1000 # for triple
+    #df.fillna(9999, inplace =True)
+    df['dtus3'] = df['dtus3'].astype('float32') #
+    dtusmin3 = df['dtus3'].min()
+    dtusmax3 = df['dtus3'].max()
+    df.fillna(99999, inplace =True)
+    print(f"D... range of time differences (erlang) values : {dtusmin3:.3f} us ...  {dtusmax3:.3f}  usec. TRIPLE" )
+
+    df['dtus4'] = (df.time.shift(-3)-df.time)*1000*1000 # for triple
+    #df.fillna(9999, inplace =True)
+    df['dtus4'] = df['dtus4'].astype('float32') #
+    dtusmin4 = df['dtus4'].min()
+    dtusmax4 = df['dtus4'].max()
+    df.fillna(99999, inplace =True)
+    print(f"D... range of time differences (erlang) values : {dtusmin4:.3f} us ...  {dtusmax4:.3f}  usec. QUADRUPLE" )
+
+    df['dtus5'] = (df.time.shift(-4)-df.time)*1000*1000 # for triple
+    #df.fillna(9999, inplace =True)
+    df['dtus5'] = df['dtus5'].astype('float32') #
+    dtusmin5 = df['dtus5'].min()
+    dtusmax5 = df['dtus5'].max()
+    df.fillna(99999, inplace =True)
+    print(f"D... range of time differences (erlang) values : {dtusmin5:.3f} us ...  {dtusmax5:.3f}  usec. PENTUPLE" )
+
+
+
 
     print(f"D... broadening table - next_E")
     df['next_E'] = (df.E.shift(-1))
     df.fillna(0, inplace=True)
     df['next_E'] = df['next_E'].astype('int32') #
 
+    print(f"D... broadening table - next_E3")  # for triple
+    df['next_E3'] = (df.E.shift(-2))
+    df.fillna(0, inplace=True)
+    df['next_E3'] = df['next_E3'].astype('int32') #
+
+    print(f"D... broadening table - next_E4")  # for triple
+    df['next_E4'] = (df.E.shift(-3))
+    df.fillna(0, inplace=True)
+    df['next_E4'] = df['next_E4'].astype('int32') #
+
+    print(f"D... broadening table - next_E5")  # for triple
+    df['next_E5'] = (df.E.shift(-4))
+    df.fillna(0, inplace=True)
+    df['next_E5'] = df['next_E5'].astype('int32') #
+
+
+
+
     print(f"D... broadening table - next_ch")
     df['next_ch'] = (df.ch.shift(-1))
     df.fillna(0, inplace=True)
     df['next_ch'] = df['next_ch'].astype('int32') #
 
     if 'extras' in df.columns:
-        print(f"D... broadening table - prev_dtus SATU")
+        print(f"D... broadening table - prev_dtus SATU  previous time")
         df['prev_satu'] = (df.time - df.time.shift(+1))*1000*1000
         #  this says - if no saturation flag=> set 0
         df['prev_satu'] = np.where((df['extras']&1)==1, df['prev_satu'] , 0)
@@ -298,9 +363,15 @@ def enhance_by_dtus_and_next_E(df):
         df.fillna(0, inplace = True)
         df['prev_satu'] = df['prev_satu'].astype('float32') #
 
-    print(df)
+    #----------nice debug print ---------------------
+    #print(df)
     endblock()
-    return df
+    return df, dtusmin, dtusmax
+
+
+
+
+
 
 #----------------------------------------------------------------------
 def pd_detect_zeroes(df, channel,  TIMEWIN_US = 4.2 ):
@@ -312,58 +383,230 @@ def pd_detect_zeroes(df, channel,  TIMEWIN_US = 4.2 ):
       2nd phase - SZ and DZ - make one event from these PILEUPs
       3rd phase - create
     """
-    print( f"D.. detecting zeroes: WINDOW == {TIMEWIN_US} us")
-    print( freemem() )
+    print( f"D... detection of zeroes: WINDOW == {TIMEWIN_US} us")
+    #print( freemem() )
 
     # this oneliner doesnt occupy memory ---------------
     # print("D... searching for correlations")
 
+    # I PREPARE SEVERAL VIEWS : only this channel :
     dfz = df[:-1]
+    index_names = dfz[  (dfz.ch!=channel)].index
+    dfz = dfz.drop(  index_names )
+    dfnomu = df # drop the close (cluster) events to count standalones
+    index_names = dfnomu[ (dfnomu.ch!=channel)].index
+    dfnomu = dfnomu.drop( index_names )
 
-    dzeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E==0)) & (dfz.dtus<TIMEWIN_US) ])
-    print(freemem() )
+    zeroes = len(  dfz.loc[ (dfz.E==0) ] )
+    print( f"TOTAL ZEROES {zeroes}")
 
-    # tzeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E==0)) & (dfz.dtus<TIMEWIN_US) ])
+    # COMBINED ZEROES
+    # ilogic zeroes
+    izeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E!=0)) & (dfz.dtus<=TIMEWIN_US) ])
+    print( "ilogiczeroes ", izeroes )
+    #izeroes = len(df_iz)
+    #print(freemem() )
+    # single zeroes
+    szeroes = len(dfz.loc[ ((dfz.E!=0) & (dfz.next_E==0)) & (dfz.dtus<=TIMEWIN_US) ])
+    print( f"singlezeroes = {szeroes} " )
+
+
+    # STANDAL EVENTS AND LATER ALSO STANDAL ZEROES
+    index_names = dfnomu[ (dfnomu.dtus<=TIMEWIN_US) | (dfnomu.dtuspr<=TIMEWIN_US)  ].index
+    dfnomu = dfnomu.drop( index_names) # NOT inplace, else it changes the df!!!
+    #index_names = dfnomu[ (dfnomu.dtuspr<TIMEWIN_US)  ].index
+    ###index_names = dfnomu[ dfnomu.dtuspr<TIMEWIN_US ].index
+    #dfnomu = dfnomu.drop( index_names ) # NOT inplace, else it changes the df!!!
+    #print(dfnomu)
+
+    print(f"i... I dropped dtus,  remains: len==",len(dfnomu) )
+    stazeroes = len( dfnomu.loc[ (dfnomu.E==0)  ])
+    print(f" ... of which {stazeroes} is standalone zeroes")
+
+
+    # playing on clusters----------------------------
+    # I drop all nonzero ene and all that are further apart (both sides)
+
+    # this would be more consistently defing the window - each subseq in in WIN
+
+    # ----------- COMPLEMENT TO STANDALONES
+    dfz = dfz[~dfz.isin(dfnomu) ].dropna()
+
+    #dfz.reset_index(inplace=True)
+
+
+    #index_names = dfz[ (dfz.dtus>=TIMEWIN_US) | (dfz.dtuspr>=TIMEWIN_US) ].index  # drop those far
+    #dfz = dfz.drop( index_names )
+    print(f"dfz (clusters) with all ene      = {len(dfz)}")
+    index_names = dfz[ (dfz.E!=0) ].index # drop nonzeroes
+    dfz = dfz.drop( index_names )
+    print(f"dfz (clusters) with  0  ene only = {len(dfz)}")
+
+    #----------------------- HERE I HAVE PURE ZERO------------------------AND CLUSTERS
+    #OK#print(dfz) # here I have all the zeroes (2*dblzero)  11,12; 20,21;
+
+
+    # I can try to count dblroes
+    # this is the start of the cluster and limited to  double clusters
+
+
+    #dfz.reset_index(inplace=True) # for any games, I need good index
+
+    index_of_doubles = ((dfz.E==0) & (dfz.next_E==0)) & (dfz.dtus<=TIMEWIN_US)& (dfz.dtuspr>TIMEWIN_US) & ( (dfz.next_E3!=0) |  (dfz.dtus3-dfz.dtus>TIMEWIN_US)  )
+    dzeroes = len( dfz.loc[ index_of_doubles ])
+
+    print( f"real doublezeroes = {dzeroes}  len of the (zero only) dfz={len(dfz)};  remains to explain {len(dfz)-2*dzeroes} zeroes" )
+# 2*dzeroes is not here, since the 2nd part of dbl is not present in in dataframe?/??
+
+    totclusters = len( dfz.loc[ (dfz.dtuspr>TIMEWIN_US)   ])
+    print( f"total clusters = {totclusters} ;  number of multiclusters (3+): {totclusters-dzeroes}; exp:{(len(dfz)-2*dzeroes)/(totclusters-dzeroes):.2f}-clusters in average (?!?)" )
+
+
+
+
+
+    print("______________________ cluster statistics _____ tag and cumulative sum_________")
+    dfz['cluster'] = 0
+    #dfz.cluster.loc[  (dfz['dtuspr']>TIMEWIN_US) ] = 1 # chaining !!!! not good
+    dfz.loc[  (dfz['dtuspr']>TIMEWIN_US) , 'cluster' ] = 1
+    #print(dfz)
+
+    dfz['clusum'] = dfz.cluster.cumsum()
+
+    #dfz['dIndex'] = dfz.index
+    #dfz.reset_index(inplace=True)
+    #dfz['dIndex'] = dfz['dIndex'] - dfz.index
+    #print(dfz)
+    print(dfz.clusum.value_counts().sort_values(ascending=False).value_counts().sort_index(ascending=False) )
+    #print(dfz.dIndex.value_counts().sort_values(ascending=False).value_counts().sort_index(ascending=False) )
+    print("______________________ cluster statistics ________________END")
+
+
+
+
+    # print(" ... dropping doublezeroes")
+
+    # index_names = dfz[ ((dfz.E==0) & (dfz.next_E==0)) & (dfz.dtus<=TIMEWIN_US)& (dfz.dtuspr>TIMEWIN_US) & ( (dfz.next_E3!=0) |  (dfz.dtus3-dfz.dtus>TIMEWIN_US)  ) ].index # drop nonzeroes
+
+    # #print(dfz.loc[ index_names] ) # here, only 11, 20, 23 remained.
+    # #print(" ... dropping", index_names," = index_names")
+
+    # #slow didnt work for i in index_names: # i need to append Index object
+    # last_index = dfz.last_valid_index() # define before, else toooo slow
+
+    # second_index = pd.Index( [ i+1 for i in index_names if i+1<=last_index] )
+
+    # #print(" ... dropping", second_index," = index_names")
+
+    # ###dfz = dfz.drop( dfz[index_of_doubles].index )
+    # dfz = dfz.drop( index_names )
+    # dfz = dfz.drop( second_index )
+    # #pd.set_option('display.max_rows', 100)
+    # #print( dfz.tail(100) )
+
+
+    # totclusters2 = len( dfz.loc[ (dfz.dtuspr>TIMEWIN_US)   ])
+    # print( f"total 3+clusters = {totclusters2} ;   exp:{len(dfz)/(totclusters2):.2f}-clusters in average" )
+
+
+
+
+
+
+
+    # # # triple zeroes - time window questionable
+    # # tzeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E==0)& (dfz.next_E3==0)) & (dfz.dtus3<2*TIMEWIN_US)& (dfz.dtuspr>TIMEWIN_US) ])
+    # # qzeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E==0)& (dfz.next_E3==0)& (dfz.next_E4==0)) & (dfz.dtus4<3*TIMEWIN_US)& (dfz.dtuspr>TIMEWIN_US) ])
+    # # # print(freemem() )
+    # # pzeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E==0)& (dfz.next_E3==0)& (dfz.next_E4==0)& (dfz.next_E5==0)) & (dfz.dtus5<4*TIMEWIN_US)& (dfz.dtuspr>TIMEWIN_US) ])
+    # # # print(freemem() )
+
+
+    # tzeroes = len( dfz.loc[ (dfz.dtuspr>TIMEWIN_US) & (dfz.dtus3-dfz.dtus<=TIMEWIN_US) & (dfz.dtus4-dfz.dtus3>TIMEWIN_US) ] )
+
+    # index_names = dfz[ (dfz.dtuspr>TIMEWIN_US) & (dfz.dtus3-dfz.dtus<=TIMEWIN_US) & (dfz.dtus4-dfz.dtus3>TIMEWIN_US)  ].index # drop triplets
+
+
+
+    # print(" ... dropping triplezeroes")
+
+
+    # #print(" ... dropping", index_names," = index_names")
+    # second_index = pd.Index( [ i+1 for i in index_names if i+1<=last_index] )
+    # third_index = pd.Index( [ i+2 for i in index_names if i+2<=last_index] )
+
+
+    # ###dfz = dfz.drop( dfz[index_of_doubles].index )
+    # dfz = dfz.drop( index_names )
+    # #print(" ... dropping", second_index," = index_names")
+    # dfz = dfz.drop( second_index )
+    # #print(" ... dropping", third_index," = index_names")
+    # dfz = dfz.drop( third_index )
+
+    # #print(dfz)
+
+    # totclusters3 = len( dfz.loc[ (dfz.dtuspr>TIMEWIN_US)   ])
+    # print( f"total 4+clusters = {totclusters3} ;   exp:{len(dfz)/(totclusters3):.2f}-clusters in average" )
+
+    # #DFszeroes =  dfz.loc[ ((dfz.E!=0) & (dfz.next_E==0)) & (dfz.dtus<TIMEWIN_US)  & (dfz.ch==channel)]
+    # #print( "singlezeroes \n", DFszeroes )
+
+    # #szeroes = len(df_sz)
+    # #print(freemem() )
+
+
+    # print("______________________ cluster statistics _____i play on sparcity now_________")
+    # dfz['cluster'] = 0
+    # #dfz.cluster.loc[  (dfz['dtuspr']>TIMEWIN_US) ] = 1 # chaining !!!! not good
+    # dfz.loc[  (dfz['dtuspr']>TIMEWIN_US) , 'cluster' ] = 1
+    # dfz['clusum'] = dfz.cluster.cumsum()
+
+    # #dfz['dIndex'] = dfz.index
+    # #dfz.reset_index(inplace=True)
+    # #dfz['dIndex'] = dfz['dIndex'] - dfz.index
+    # print(dfz)
+    # print(dfz.clusum.value_counts().sort_values(ascending=False).value_counts().sort_index(ascending=False) )
+    # #print(dfz.dIndex.value_counts().sort_values(ascending=False).value_counts().sort_index(ascending=False) )
+    # print("______________________ cluster statistics ________________END")
+
+
+    # #zeroes2 = len( dfz[ (dfz.E==0) & (dfz.ch==channel)] )
+    # #zeroes = len(df_z)
     # print(freemem() )
 
-    szeroes = len( dfz.loc[ ((dfz.E!=0) & (dfz.next_E==0)) & (dfz.dtus<TIMEWIN_US) ] )
-
-    DFszeroes =  dfz.loc[ ((dfz.E!=0) & (dfz.next_E==0)) & (dfz.dtus<TIMEWIN_US) ]
-    print( DFszeroes )
-
-    #szeroes = len(df_sz)
-    print(freemem() )
-
-    izeroes = len( dfz.loc[ ((dfz.E==0) & (dfz.next_E!=0)) & (dfz.dtus<TIMEWIN_US) ] )
-    #izeroes = len(df_iz)
-    print(freemem() )
 
 
-    print(f"D... ---------------------- START of zero detection ch={channel}")
-    zeroes = len( dfz.loc[ (dfz.E==0) & (dfz.ch==channel)] )
-    zeroes2 = len( dfz[ (dfz.E==0) & (dfz.ch==channel)] )
-    #zeroes = len(df_z)
-    print(freemem() )
 
+    #    D... of which is separated  = {len(dfnomu)-stazeroes:8d}
+    output = f"""
 
-    print(f"D... total  zeroes (chan={channel}) = {zeroes} which is  {zeroes/len(df)*100:5.2f}%")
-    print(f"D... total  zeroes (chan={channel}) = {zeroes2} which is  {zeroes2/len(df)*100:5.2f}%")
-
-    #print(f"D... triple zeroes (chan={channel}) = {tzeroes} which is  {tzeroes/len(df)*100:5.2f}%")
-
-    print(f"D... double zeroes (chan={channel}) = {dzeroes} which is  {dzeroes/len(df)*100:5.2f}%")
-    #print(df_dz)
-    print(f"D... single zeroes (chan={channel}) = {szeroes} which is  {szeroes/len(df)*100:5.2f}%")
-    #print(df_sz)
-    print(f"D... ilogic zeroes (chan={channel}) = {izeroes} which is  {izeroes/len(df)*100:5.2f}%")
-    #print(df_iz)
-    print("D...  There REMAINS ",  zeroes -  dzeroes*2," zeroes that are NOT DOUBLE"  )
-    print("D... ______________________ end of zero detection")
+    D... total  events (chan={channel}) = {len(df):8d}
+    D... total  zeroes (chan={channel}) = {zeroes:8d}   ~  {zeroes/len(df)*100:5.2f}%
+    D... total  nonzero(chan={channel}) = {len(df)-zeroes:8d}
 
+    D... standl.zeroes (chan={channel}) = {stazeroes:8d} ~  {stazeroes/len(df)*100:5.2f}%
+    D... => zrs@clusters        = {zeroes-stazeroes:8d}    (zeroes that must be in clusters)
 
+    D... double zeroes (chan={channel}) = {dzeroes:8d} ~  {2*dzeroes/len(df)*100:5.2f}% ( % counted correctly 2x)
+    D... => zrs@HigherClusters  = {zeroes-stazeroes-2*dzeroes:8d}  (zeroes in longer than 2 clusters)
+    D... single zeroes (chan={channel}) = {szeroes:8d} ~  {szeroes/len(df)*100:5.2f}%
+    D... ilogic zeroes (chan={channel}) = {izeroes:8d} ~  {izeroes/len(df)*100:5.2f}%
+    D... ______________________ end of zero detection
+    D...  window {TIMEWIN_US} us   ....  crutial for correct double zeroes detection
+"""
     # returns length for now
     endblock()
-    return zeroes,dzeroes,szeroes,izeroes
+    return zeroes,dzeroes,szeroes,izeroes,stazeroes,output
+
+
+
+
+
+
+
+
+
+
 
 #------------------------------------------------------------------------
 
@@ -392,9 +635,9 @@ register 0x1n80 to enable the fake-event saving in case of internal roll-over. I
 will have both bit[3] and bit[2] set to 1, while in the second case the event will have both bit[3] and
 bit[1] set to 1.
     """
-    print(f"D... fourcolumns=={fourcolumns}")
-    print(f"D... fourcolumns=={fourcolumns}")
-    print(f"D... fourcolumns=={fourcolumns}")
+    #print(f"D... fourcolumns=={fourcolumns}")
+    #print(f"D... fourcolumns=={fourcolumns}")
+    #print(f"D... fourcolumns=={fourcolumns}")
     start = dt.datetime.now()
     df = pd.read_table( filename,
                         names=['time','E','pu','ch','extras'],
@@ -402,18 +645,18 @@ bit[1] set to 1.
                         comment="#")
 
     meastime = (dt.datetime.now() - start).total_seconds()
-    print(f"D... speed = {get_number_of_lines(filename)/meastime/1000/1000:.3f} Mrows/sec")
+    print(f"D... table is in memory ...  {get_number_of_lines(filename)/meastime/1000/1000:.3f} Mrows/sec")
 
-    print(f"D... fourcolumns=={fourcolumns}")
+    #print(f"D... fourcolumns=={fourcolumns}")
 
-    print("D... lasttime ",df.iloc[-1]["time"])
-    print(freemem() )
+    #print("D... table last timestamp ",df.iloc[-1]["time"]/1e+8, "sec.")
+    print( freemem() )
     # drop if 4 column
     df = df.dropna(axis='columns', how='all')
 
 
     df['time'] = df['time']/1e+8 # this is correct for 10ns step ADC
-    print("D... lasttime ",df.iloc[-1]["time"])
+    print("D... table last timestamp ",df.iloc[-1]["time"], "sec.")
 
     #print(df.dtypes)
     df['time'] = df['time'].astype('float64')
@@ -422,11 +665,11 @@ bit[1] set to 1.
     df['ch']   = df['ch'].astype('int32')
     if 'extras' in df:
         df['extras'] = df['extras'].astype('int32')
-    print("D...    finale types ************************************")
-    print(df.dtypes)
+    #print("D...    finale types ************************************")
+    #print(df.dtypes)
 
     if fourcolumns:
-       print("D... extra operation for four columns ... ")
+       print("D... extra operation for four columns ...  FOURCOLUMNS TABLE !!!")
        print(df)
        df['E'] = df['E']+ (df['pu']&1) * 16384
        df['pu'] = df['pu'].apply(lambda x: x >> 1)
@@ -449,7 +692,7 @@ bit[1] set to 1.
         #print(df)
         #print( freemem() )
 
-    print("D... lasttime ",df.iloc[-1]["time"])
+    print("D... table last timestamp ",df.iloc[-1]["time"], "sec.")
     endblock()
     return df
 
